from pysqlite2 import dbapi2 as sqlite
import time, os
from math import sqrt
from datetime import date, datetime, timedelta
import sys, os.path, getopt, ConfigParser

 
def create_db_table( con, cursor ):
    " Creates the db and the db-table "

    global db_table_name, debug
   
    cursor.execute('CREATE TABLE %s \
        (logdate DATE, pattern TEXT, count INTEGER, PRIMARY KEY (logdate,pattern))' % db_table_name )
    con.commit()

    if debug: print "Table '%s' successfully created." % db_table_name

def db_cleanup( cursor, con ) :
    " Deletes old db records "

    global db_table_name, logdate, db_prune_threshold, debug

    td = timedelta( db_prune_threshold )
    # oldest allowed date to reside in db:
    allowed_date = logdate - td
    if debug: print "Pruning away pattern data older than %s." % allowed_date
    
    cursor.execute("DELETE from %s where logdate < '%s'" % (db_table_name, allowed_date ) )
    con.commit()

def read_db_patterns( cursor, con ):
    " Fetches all patterns stored in db "
    
    global db_table_name, debug

    db_patterns = {}

    cursor.execute("SELECT * from %s" % ( db_table_name ))
    
    for ( logdate, pattern, count ) in cursor:
        if pattern not in db_patterns :
            db_patterns[pattern] = {}
        # logdate can only occur once per pattern, see primary key
        db_patterns[pattern][logdate] = count
   
    return db_patterns

def frequent_db_patterns( cursor, con, patterns, db_patterns ) :
    " Returns frequent db patterns not represented among the patterns "
    
    global frequent_pattern_threshold, db_table_name, debug
    
    cursor.execute("SELECT COUNT(distinct logdate) from %s" % ( db_table_name ))

    no_unique_dates = cursor.fetchone()[0] 
    if debug: print "\tThere are %i unique dates in db" % no_unique_dates

    frequent_db_patterns = {}

    # Find all db keys not present in patterns
    keyset = set( db_patterns.keys() ) - set( patterns.keys() )

    for key in keyset:
        dict = db_patterns[key]
        no_instances = len( dict )
        if ( float( no_instances )/no_unique_dates ) >= frequent_pattern_threshold :
            if debug: 
                print "\tDB pattern '%s' has freq %.2f. Considered as frequent, but missing." % ( key, float( no_instances )/no_unique_dates )
            tot = sum( dict.values() )
            avg = float( tot )/no_instances
            frequent_db_patterns[key] = avg
    
    return frequent_db_patterns


def read_patterns( file ):
    " Reads generated patterns from file and into a dictionary "

    if debug: print "Reading patterns ..."
    
    patterns = {}

    f = open(file, 'r')

    for line in f:
        line = line.split()
        count = int( line.pop() )
        patterns[' '.join(line)] = count
    f.close()

    return patterns

def save_patterns( patterns, con, cursor ):
    " Insert all patterns from the patterns file into the database "
  
    global db_table_name, logdate, reportfile, debug

    if debug: print "Saving patterns(%i) to database." % len(patterns)
    
    try:
        cursor.execute('begin;')
        for pattern, count in patterns.items() :
            cursor.execute("insert into %s (logdate, pattern, count) \
                values (?,?,?)" % (db_table_name) , ( logdate, pattern, count))
        con.commit()
    except:
        print "Error: "
        print "\tSomething went wrong. Could not save patterns."
        print "\tNo patterns saved to database, and no report generated. Quitting." 
        con.close()
        sys.exit()
        

    f = open(reportfile, 'w')
    f.writelines("%i patterns saved to database." % len(patterns)) 
    f.close()

def analyse_patterns( patterns, db_patterns ) :
    " Compares patterns to patterns in database "

    global reportfile, standard_deviations, debug

    if debug: "Analysing patterns ..."

    new_patterns = {}
    underrepresented = {}
    overrepresented = {}

    p_set = set( patterns.keys() )
    db_set = set( db_patterns.keys() )

    new_keys = p_set - db_set
    for key in new_keys :
        new_patterns[key] = patterns[key]

    similar = p_set.intersection( db_set )
    for pattern in similar :
        dbp = db_patterns[pattern]
        no_instances = len( dbp )
        total_val = sum( dbp.values() )
        avg = float( total_val ) /no_instances 

        squared_total = 0
        for val in dbp.values() :
            squared_total += ( ( val - avg )**2 )
        
        # Standard deviation
        sd = sqrt( squared_total/no_instances ) 

        # Investigate the corresponding pattern in the pattern dict
        count = patterns[pattern]

        if count > ( avg + ( standard_deviations * sd ) ) :
            overrepresented[pattern] = {'count':count, 'mean':avg, 'sd':sd}
        elif count < ( avg - ( standard_deviations * sd ) ) :
            underrepresented[pattern] = {'count':count, 'mean':avg, 'sd':sd}
    
    return new_patterns, underrepresented, overrepresented


def write_report( missing_patterns, new_patterns, underrepresented, overrepresented ) :
    " Write analysis report to report file "

    global reportfile, debug 

    if debug: print "Writing report ..."

    f = open(reportfile, 'a')
    
    f.writelines("\nNew: \n")  
    if not new_patterns :
        f.writelines("\tNo new patterns\n")
    for pattern, count in new_patterns.items():
        f.writelines("\t%10d,\t\t%s\n" % (count, pattern))

    f.writelines("Overrepresented: \n")
    if not overrepresented :
        f.writelines("\tNo patterns\n")
    for pattern, dict in overrepresented.items() :
        count = dict['count']
        avg = dict['mean']
        sd = dict['sd']
        f.writelines("\t%10d ( mean: %14.2f,\tsd: %5.2f )  %s\n" % (count, avg, sd, pattern))        
    
    f.writelines("Underrepresented: \n")
    if not underrepresented :
        f.writelines("\tNo patterns\n")
    for pattern, dict in underrepresented.items() :
        count = dict['count']
        avg = dict['mean']
        sd = dict['sd']
        f.writelines("\t%10d ( mean: %14.2f,\tsd: %5.2f )  %s\n" % (count, avg, sd, pattern))        


    f.writelines("Missing: (mean count)\n")
    if not missing_patterns :
        f.writelines("\tNo missing frequent patterns")
    for pattern, avg in missing_patterns.items() :
        f.writelines("\t%10d,\t\t%s\n" % (avg, pattern))
   
    f.close()

def main() :
    global logdate, logfile, patternfile, reportfile

    global debug 
    global db_file, db_table_name

    db_name = os.path.basename(db_file)

    # If db does not already exist, create it first.
    create = False

    if not os.path.isfile(db_file):
        create = True
        if debug: print "Will create db '%s'..." % db_file

    con = sqlite.connect(db_file)
    cursor = con.cursor()

    # Only run when createing the db for the first time. Dah ...
    if create:
        if debug: print "Creating databse '%s'..." % ( db_name )
        create_db_table( con, cursor )
    else:
        cursor.execute("SELECT COUNT(*) from sqlite_master where name=?", (db_table_name ,))
        if cursor.fetchone()[0] == 0 :
            create_db_table( con, cursor )
            if debug: print "New table '%s' created." % db_table_name

        cursor.execute("SELECT COUNT(distinct logdate) from %s where logdate = '%s'" % ( db_table_name, logdate ))
        if cursor.fetchone()[0] != 0 :
            print "Error: "
            print "\tLogdate %s already exists in database. Quitting." % logdate
            print "\tNo patterns saved to database, and no report generated. Quitting." 
            con.close()
            sys.exit()

        # Clean up db and read db records
        db_cleanup( cursor, con )
        
        # Read patterns from database
        db_patterns = read_db_patterns( cursor, con )
    
    # Read patterns from pattern file
    patterns = read_patterns( patternfile )
    
    # If db not new and db contains patterns, analyse results
    if ( not create ) and ( db_patterns ) : 
        missing_patterns = frequent_db_patterns( cursor, con, patterns, db_patterns )
        new_patterns, underrepresented, overrepresented = analyse_patterns( patterns, db_patterns )

    # Save patterns
    if not patterns:
        print "Error:"
        print "No patterns found. Exiting."
        sys.exit()

    save_patterns( patterns, con, cursor )

    con.close()

    if ( not create ) and ( db_patterns ) :
        # Do not write report until the patterns are safely stored in db
        write_report( missing_patterns, new_patterns, underrepresented, overrepresented )

    if debug: print "Finished!"


def usage() :
    print "\nUsage of patter mining-based log analyser: "
    print "\t-h\t(--help) To show this message."
    print "\t-p\tFull path to the generated patterns (mandatory)."
    print "\t-r\tFull path to the directory where the results shoul be written (mandatory)."
    print "\t-t\tLog date on the form 'yyyy-mm-dd'. The default is today's date"
    print "\t-l\tFull path to the logfile that generated the patterns."
    print "\t-c\t(--config) Full path to a different config file, the default is 'statistical.config'."
    print "\t-d\t(--debug) To turn on debugging code."


if __name__ == "__main__":
   
    debug = False 
    logdate = date.today()
    patterfile = None
    report_name = 'report'
    reportfile = None
    config_file = 'statistical.config' 

    argv = sys.argv[1:]

    try:
        opts, args = getopt.getopt(argv, "hdc:r:p:t:l:", ["help", "debug", "config="]) 
    except getopt.GetoptError, e:
        print e
        usage()
        sys.exit(2)
    if ( '-h' in argv ) or ( '--help' in argv ) :
        usage()
        sys.exit()
    if ( '-p' not in argv ) or ( '-r' not in argv ) :
        usage()
        sys.exit(2)

    for opt, arg in opts :
        if opt in ('-c','--config') :
            if not os.path.exists( arg ) :
                print "Config file '%s' does not exist." % arg
                sys.exit()
            print "Using config file: %s" % arg
            config_file = arg
        elif opt in ('-d', '--debug' ) :
            debug = True
        elif opt == '-t' :
            try:
                logdate = datetime.strptime(arg,'%Y-%m-%d').date()
            except:
                print "Log date must be on the form yyyy-mm-dd."
                sys.exit()
        elif opt == '-l' :
            if not os.path.isfile( arg ) :
                print "Logfile '%s' does not exist." % arg
                sys.exit()
            logfile = arg
        elif opt == '-p' :
            if not os.path.isfile( arg ) :
                print "Pattern file '%s' does not exist." % arg
                sys.exit()
            patternfile = arg
        elif opt == '-r' :
            if not os.path.isdir( arg ) :
                print "'%s' is not a valid directory." % arg
                sys.exit()
            path = arg
            if arg[-1] != '/':
                path = arg+'/'

    # Must make sure that the logdate has been registered before we set
    # the path to the reportfile
    reportfile = "%s%s_%s" % ( path, logdate, report_name )
    
    # Read config file

    config = ConfigParser.RawConfigParser()
    config.read( config_file )

    db_file                     = config.get('DB', 'db_file')
    db_table_name               = config.get('DB', 'db_table_name')
    db_prune_threshold          = config.getint('DB', 'db_prune_threshold')
    frequent_pattern_threshold  = config.getfloat('Pattern_analysis', 'frequent_pattern_threshold')
    standard_deviations         = config.getint('Pattern_analysis', 'standard_deviations')


    main()
